from bolib3 import np

"""
Reference
    John F. Bard
    Practical Bilevel Optimization: Algorithms and Applications
    Chapter 7, example 7.2.1, page 275
"""

# Properties
name: str = 'bard721'
category: str = 'textbook'
subcategory: str = 'bard_1998'
datasets: list = []

# Feasible point
x0 = np.array([0.8, 0.2])
y0 = np.array([0.0])


# Methods
def F(x, y, data=None):
    """
    Upper-level objective function
    (quadratic)
    """
    return 0.5*(x[0] - 0.8)**2 + 0.5*(x[1] - 0.2)**2 + 0.5*(y[0] - 1.0)**2


def G(x, y, data=None):
    """
    Upper-level inequality constraints
    (bounds)
    """
    return np.array([
        x[0],
        x[1],
        1-x[0],
        1-x[1]
    ])


def H(x, y, data=None):
    """
    Upper-level equality constraints
    (none)
    """
    return np.empty(0)


def f(x, y, data=None):
    """
    Lower-level objective function
    (quadratic)
    """
    return 0.5*y[0]**2 + y[0] - x[0]*y[0] + 2*x[1]*y[0]


def g(x, y, data=None):
    """
    Lower-level inequality constraints
    (bounds)
    0 <= y <= 1
    """
    return np.array([
        y[0],
        1-y[0]
    ])


def h(x, y, data=None):
    """
    Lower-level equality constraints
    (none)
    """
    return np.empty(0)


def read_data(filepath=''):
    """
    If the bilevel program is parameterized by data, this function should
     provide code to read data file and return an appropriate python structure.
    """
    pass


def dimensions(key='', data=None):
    """
    If the argument 'key' is not specified, then:
     - a dictionary mapping variable/function names (str) to the corresponding dimension (int) is returned.
    If the first argument 'key' is specified, then:
     - a single integer representing the dimension of the variable/function with the name {key} is returned.
    """
    n = {
        "x": 2,  # Upper-level variables
        "y": 1,  # Lower-level variables
        "F": 1,  # Upper-level objective functions
        "G": 4,  # Upper-level inequality constraints
        "H": 0,  # Upper-level equality constraints
        "f": 1,  # Lower-level objective functions
        "g": 2,  # Lower-level inequality constraints
        "h": 0   # Lower-level equality constraints
    }
    if key in n:
        return n[key]
    return n
